import java.io.*;
import java.util.Scanner;

/**
 * Assembler-Parser.
 *
 * @author Owyn
 * @version 2008-05-08
 */
public class Parser {

    // Befehle als Strings
    static final String[] opcodes = {
    "ADD","SUB","MUL","DIV","MOD","CMP","NOT","AND","OR","XOR","SH",
    "ADDI","SUBI","MULI","DIVI","MODI","CMPI","ANDI","ORI","XORI","SHI",
    "LDO","LDX","STO","STX",
    "BEQ","BNE","BLT","BGE","BLE","BGT","JSR","RET",
    "RDD","WRD","WRH","WRL","HALT"};

    // Befehle als OpCodes
    static final int ADD  = 0;  // Ra = Rb + Rc
    static final int SUB  = 1;  // Ra = Rb - Rc
    static final int MUL  = 2;  // Ra = Rb * Rc
    static final int DIV  = 3;  // Ra = Rb / Rc
    static final int MOD  = 4;  // Ra = Rb % Rc
    static final int CMP  = 5;  // Ra = SIGNUM(Rb - Rc)

    static final int NOT  = 6;  // Ra = NOT Rb = ~Rb
    static final int AND  = 7;  // Ra = Rb AND Rc = Rb & Rc
    static final int OR   = 8;  // Ra = Rb OR Rc = Rb | Rc
    static final int XOR  = 9;  // Ra = Rb XOR Rc = Rb ^ Rc
    static final int SH   = 10; // ARITHMETIC SHIFT: Ra = Rb SHIFTED BY Rc

    static final int ADDI = 11; // Ra = Rb + c
    static final int SUBI = 12; // Ra = Rb - c
    static final int MULI = 13; // Ra = Rb * c
    static final int DIVI = 14; // Ra = Rb / c
    static final int MODI = 15; // Ra = Rb % c
    static final int CMPI = 16; // Ra = SIGNUM(Rb - c)

    static final int ANDI = 17; // Ra = Rb AND c = Rb & c
    static final int ORI  = 18; // Ra = Rb OR c = Rb | c
    static final int XORI = 19; // Ra = Rb XOR c = Rb ^ c
    static final int SHI  = 20; // ARITHMETIC SHIFT: Ra = Rb SHIFTED BY c

    static final int LDO  = 21; // LOAD +OFFSET: Ra = MEM[Rb + c]
    static final int LDX  = 22; // LOAD +INDEX: Ra = MEM[Rb + Rc]
    static final int STO  = 23; // STORE +OFFSET: MEM[Rb + c] = Ra
    static final int STX  = 24; // STORE +INDEX: MEM[Rb + Rc] = Ra

    static final int BEQ  = 25; // BRANCH TO INSTRUCTION +/- c IF Ra IS EQUAL TO ZERO
    static final int BNE  = 26; // BRANCH TO INSTRUCTION +/- c IF Ra IS NOT EQUAL TO ZERO
    static final int BLT  = 27; // BRANCH TO INSTRUCTION +/- c IF Ra IS LESS THAN ZERO
    static final int BGE  = 28; // BRANCH TO INSTRUCTION +/- c IF Ra IS GREATER THAN OR EQUAL TO ZERO
    static final int BLE  = 29; // BRANCH TO INSTRUCTION +/- c IF Ra IS LESS THAN OR EQUAL TO ZERO
    static final int BGT  = 30; // BRANCH TO INSTRUCTION +/- c IF Ra IS GREATER THAN ZERO
    static final int JSR  = 31; // JUMP TO INSTRUCTION c, SAVE PC+1 IN REGISTER 31
    static final int RET  = 32; // JUMP TO INSTRUCTION Rc, HALT IF c = 0

    static final int RDD  = 33; // READ DECIMAL NUMBER FROM INPUT INTO Ra
    static final int WRD  = 34; // WRITE DECIMAL NUMBER FROM Ra TO OUTPUT
    static final int WRH  = 35; // WRITE NUMBER FROM Ra TO OUTPUT IN HEXADECIMAL NOTATION
    static final int WRL  = 36; // WRITE NEWLINE TO OUTPUT

    static final int HALT = 37; // HALT

    /**
     * Liest Assembler-Code in Programmspeicher.
     *
     * @param filename Datei mit Assembler-Code
     * @param P Programmspeicher
     */
    public static void parse(String filename, int[] P) {
        BufferedReader br = null;
        String line = "";
        String[] tokens;
        int index = 0;
        int op = -1;

        try {
            br = new BufferedReader(new FileReader(filename)); // Datei oeffnen

            line = br.readLine(); // Erste Zeile lesen
            while (line != null) { // So lange Datei nicht zu Ende
                if (line.indexOf('/') >= 0) line = line.substring(0, line.indexOf('/')); // Kommentare entfernen!
                if (!line.trim().equals("")) { // Leerzeilen ignorieren
                    line = line.trim().toUpperCase(); // Leerzeichen entfernen und in Grossbuchstaben umwandeln

                    tokens = line.split("(\\s)+"); // Zeile in Woerter aufteilen

                    // 1. Wort in Zeile ist Befehl, dieser wird in einen OpCode umgewandelt.
                    boolean opCodeRecognized = false;
                    for (int i = 0; !opCodeRecognized && i < opcodes.length; i++) {
                        if (tokens[0].equals(opcodes[i])) {
                            op = i; // Befehl gefunden
                            opCodeRecognized = true;
                        }
                    }
                    if (!opCodeRecognized) throw new Exception("Unbekannter OpCode '" + tokens[0] + "'"); // Fehler?
                    P[index] = op; // OpCode in Programmspeicher schreiben

                    for (int i = 1; i < tokens.length && i < 4; i++) { // Parameter einlesen...
                        if (tokens[i].charAt(0) == 'R') tokens[i] = tokens[i].substring(1); // ..."R" entfernen...
                        P[index + i] = Integer.parseInt(tokens[i]); // ...und als Zahl in Programmspeicher schreiben
                    }

                    index += 4; // Naechster Befehl
                }

                line = br.readLine(); // Naechste Zeile lesen
            }

        } catch (Exception e) {
            System.err.println("Fehler beim Lesen oder Oeffnen der Datei " + filename + ":");
            System.err.println(e.getMessage());
            System.exit(1);
        }

    }


    /**
     * Liest Werte aus einer Datei in Datenspeicher.
     *
     * @param filename Datei mit Werten
     * @param M Datenspeicher
     * @param offset Ab dieser Adresse in M werden die Werte gespeichert
     */
    public static void load(String filename, int[] M, int offset) {
        int index = 0;
        try {
            Scanner s = new Scanner(new File(filename)); // Datei oeffnen
            while (s.hasNext()) {
                M[offset + index++] = Integer.parseInt(s.next());
            }
        } catch (Exception e) {
            System.err.println("Fehler beim Lesen oder Oeffnen der Datei " + filename + ":");
            System.err.println(e.getMessage());
            System.exit(1);
        }

    }

}
